#ifndef UTILITIEC_TRACINGHEAP_H
#define UTILITIEC_TRACINGHEAP_H

#include "../allocator-interface/allocator-interface.h"

#include <stdalign.h>

enum TracingColor {
	TRACINGCOLOR_BLACK,
	TRACINGCOLOR_GREY,
	TRACINGCOLOR_WHITE,
};

typedef struct TracingObject {
	struct TracingObject* global_next;
	struct TracingObject* color_next;
	struct TracingObject* color_prev;
	enum TracingColor color;

	// data member of the requested size
	alignas(void*) char data[];
} TracingObject;

typedef int (*TraceReferenceCallback) (void* context, void* reference);
typedef int (*ReferenceTracingFunction) (void* context, void* object, TraceReferenceCallback trace_callback, void* callback_context);

typedef struct TracingHeap_s {
	TracingObject* objects;
	TracingObject* white_objects;
	TracingObject* grey_objects;
	TracingObject* black_objects;

	ReferenceTracingFunction tracing_function;
	void* tracing_function_context;

	enum TracingColor reachable_color;
	enum TracingColor unreachable_color;

	allocator_t* allocator;
} TracingHeap;

int TracingHeap_Create(TracingHeap* self, ReferenceTracingFunction tracing_function, allocator_t* allocator);
void TracingHeap_Destroy(TracingHeap* self);

void* TracingHeap_Allocate(TracingHeap* self, size_t bytes);

int TracingHeap_AddTracingRoot(TracingHeap* self, void* data);

#endif // header
